/***********************************************************************/
/*                                                                     */
/* RIGI C Parser --- Copyright (C) 1996-1999                           */
/*                    The Rigi Group, University of Victoria           */
/*                                                                     */
/* see the file license.terms for usage and copying policies           */
/*                                                                     */
/* Maintainer: Johannes Martin (jmartin@csr.csc.uvic.ca)               */
/*                                                                     */
/***********************************************************************/
%{

// $Author: jmartin $
// $Date: 2001/03/02 19:32:14 $
// $Source: /home/rigi/proj/cvs/Rigi/rigiparse/cparse.y,v $
// $Locker:  $
// $State: Exp $
//
// ------------------------------------------------------------------------

/* 	This grammer was supplied by (via the net)

		Nick Crossley
		Computer Consoles Inc.
		9801 Muirlands Boulevard,
		Irvine
		CA 92718
		(714) 458-7282

		email: ...!uunet!ccicpg!nick

	credit for the grammer should be directed to him.
	There has been minor minor changes made from the
	original version received.

	[MJW 94] Many more changes made, mostly to make grammar more general,
	eg. with the new non-terminal id_or_name. Plus, more flexibility
	as to where semicolons, commas, etc. can go.
*/

#include <stdlib.h>

#include "clex.h"
#include "cinit.h"
#include "symtbl.h"
#include "parse.h"
#include "parmlist.h"

// The following routines must be supplied by the program using the compiler

extern "C" int	yyparse();		// interface to Yacc Parser
//extern int     yylineno;

// Redefine YYMAXDEPTH (parse stack) to be greater than the Yacc default

#define YYINITDEPTH     500

///////////////////////////////////////////////////////////////////////////

/* if there is a prototype of the same name defined:
 * - create an arc from that prototype to the new symbol and overwrite
 *   the prototype by this new symbol
 * if there is a function of the same name defined:
 * - create an arc from the new symbol to that function
 * otherwise
 * - add a new symbol
 *
 * ParseInfo - current token info
 */
void checkaddfunc(ParseInfo &parseInfo) {
  const ParseInfo* foundInfo = findsym(parseInfo.Name());

  if (foundInfo == NULL) {
    addsym(parseInfo, true);
    return;
  }

  ParseInfo temp = *foundInfo;
  addsym(parseInfo, true);
  if (temp.Storage() == STORAGE_EXTERN)
    newprototypes(temp, parseInfo);
  else
    newprototypes(parseInfo, temp);
}

/* if there is an external variable of the same name defined:
 * - create an arc from that external variable to the new symbol and overwrite
 *   the external variable by this new symbol
 * if there is a variable of the same name defined:
 * - create an arc from the new symbol to that variable
 * otherwise
 * - add a new symbol
 *
 * ParseInfo - current token info
 */
void checkaddvar(ParseInfo &parseInfo) {
  if (parseInfo.Prototype()) {
    checkaddfunc(parseInfo);
    return;
  }
    
  const ParseInfo* foundInfo = findsym(parseInfo.Name());

  if (foundInfo == NULL) {
    addsym(parseInfo, false);
    return;
  }

  ParseInfo temp = *foundInfo;
  if (parseInfo.Storage() == STORAGE_EXTERN) {
    addsym(parseInfo, true);
    newdeclares(parseInfo, temp);
    return;
  }

  if ((temp.Storage() == STORAGE_EXTERN) && !infunc()) {
    addsym(parseInfo, true);
    newdeclares(temp, parseInfo);
    return;
  }

  addsym(parseInfo, false);
}

void appendToName(ParseInfo &info, const char *string) {
  char tmp[strlen(info.Name()) + strlen(string) + 1];
  strcpy(tmp, info.Name());
  strcat(tmp, string);
  info.SetName(tmp);
}

void voidExpression(ParseInfo &info) {
  info.SetName(NULL);
  info.SetType("void");
  info.SetPrototype(false);
}

%}

/* [MJW 94] I would like to keep the parse stack as simple as possible */

/*%union { ParseInfo info; }*/

%token	<info>	IDENTIFIER
%token  <info>	DATATYPENAME	PROTOTYPENAME	
%token	<info>	CONSTANT        STRING_CONSTANT

%token	<info>	AUTO		BREAK		CASE		CHAR
		CONTINUE	DEFAULT		DO		DOUBLE
		ELSE		ENUM		EXTERN		FLOAT
		FOR		GOTO		IF		INT
		LONG		REGISTER	RETURN		SHORT
		SIZEOF		STATIC		STRUCT		SWITCH
		TYPEDEF		TYPEOF		UNION		UNSIGNED
		VOID		WHILE		SIGNED		
		

%token		ELLIPSIS	CONST		ASM		VOLATILE

		Shift		And		Or		Rel_Op
		Eq_Op		IncDec		Asgn_Op		Arrow

%right		Asgn_Op
%right		'?'		':'	ZEB
%left		Or
%left		And
%left		'|'
%left		'^'
%left		'&'
%left		Eq_Op
%left		Rel_Op
%left		Shift
%left		'+'		'-'
%left		'*'		'/'		'%'
%left		Prefix
%left		SizeOf
%left		IncDec		Arrow		'.'
%left		'['		'('

%type   <info>   function_decl prototype_decl prototype_decl2 prototype_decl3
%type   <info>   id_or_name func_dcltr func_declarator
%type   <info>   declarator declarator2 declarator3
%type	<info>	 expr comp_expr func_expr exprs constant string_constant
%type   <info>   type_name cast_spec 
%type	<info>	 type_spec type_spec2 type_spec3 type_dcltrs 
%type   <info>   int_spec float_spec enum_spec struct_spec union_spec
%type	<info>	 exprOrInit init_list unORsigned
%type	<info>	 compound statements statement_or_label statement_no_label
%type	<info>	 expr_statement


/*	[MJW 94] I don't think I should have to do the above!!	*/


%%

/*
	[MJW 94]
	The main idea in the grammar itself is to keep it as "clean"
	as possible. Thus, there is little "action" code. Note that
	most of the latter are statements of the form "$$=$2". I
	wish some automation in the driver loop could take care of some
	of that (eg. "$$ = the first $-able thing in the RHS").

	Also note that memory allocation for <info> on the parse stack
	is not clean. Again, the parse driver should take care of freeing
	stack memory allocation.

	Of course, there's no sense worrying about the above until the
	parser does a little more. :-)

	NOTE: there is one shift/reduce conflict, which is the old
	if-else ambiguity, which is quite acceptable.
*/


/*--------------------------   DECLARATIONS   -------------------------------*/

program		:	top_level_decls
		;

top_level_decls	:
		|	top_level_decls top_level_decl
		;

top_level_decl	:	function_decl		{
                                                  $1.SetStorage(STORAGE_NONE);
						  addsym($1, true);
                                                  while (ParameterList::more())
                                                    newvar(ParameterList::get());
                                                  endfunc($1);
                                                  ParameterList::close();
                                                  exitscope();
                                                }
                |       prototype_decl          {
						  $1.SetStorage(STORAGE_EXTERN);
						  addsym($1, true);
                                                  endproto($1);
                                                  ParameterList::close();
                                                  exitscope();
                                                }
		|	var_declaration
		|       ASM ';'
		|	';'
		|	error '}'
		|	error ';'
		;

function_decl	:	specs func_dcltr var_decls compound         { $$ = $2; }
                |       specs_notype func_dcltr var_decls compound  { $$ = $2; }
		|	no_specs func_dcltr var_decls compound 	    { $$ = $2; }
		|	specs func_dcltr var_decls ASM              { $$ = $2; }
                |       specs_notype func_dcltr var_decls ASM       { $$ = $2; }
		|	no_specs func_dcltr var_decls ASM 	    { $$ = $2; }
		;

prototype_decl  :       prototype_decl2 ';'
                |       prototype_decl2 ',' 	{
						  $1.SetStorage(STORAGE_EXTERN);
						  addsym($1, true);
                                                  endproto($1);
                                                  ParameterList::close();
                                                  exitscope();
                                            	}
			  prototype_decl3 ';'	{ $$ = $4; }
                ;

prototype_decl2 :       specs func_dcltr	{ $$ = $2; }
                |       specs_notype func_dcltr	{ $$ = $2; }
                |       no_specs func_dcltr	{ $$ = $2; }
		;

prototype_decl3 :	func_dcltr
		|	prototype_decl3 ',' 	{
						  $1.SetStorage(STORAGE_EXTERN);
						  addsym($1, true);
                                                  endproto($1);
                                                  ParameterList::close();
                                                  exitscope();
                                            	}
			  func_dcltr		{ $$ = $4; }
		;

no_specs	:	/* empty */		{ setstorage(STORAGE_NONE);
						  setfunctype("int"); }
		;

func_dcltr	:   	func_declarator	{ $1.SetType(getfunctype());
					  $1.SetStorage(getstorage());
                                          checkaddfunc($1);
					  startfunc($1);
                                        } parameter_list { $$ = $1; }
		;

func_declarator	:	IDENTIFIER
                |       '*' func_declarator %prec Prefix           { $$ = $2; }
		|       '(' func_declarator parameter_list ')' 	   { $$ = $2; }
                ;

decls		:	/* empty */
		|	decls declaration
		;

asm_opt		:	/* empty */
		|	ASM
		;

declaration	:	specs declaration2_1 asm_opt ';'
                |       specs_notype declaration1_1 asm_opt ';'
		|	specs ';'
                |       specs_notype ';'
		|	linkage ';'
		|	TYPEDEF type_spec { settypedeftype($2); } type_dcltrs ';'
                |	TYPEDEF type_spec ';' 			{ ; }
		;

proto_dcltr	:	func_dcltr	{
					  $1.SetStorage(STORAGE_EXTERN);
					  addsym($1, true);
					  endproto($1);
					  ParameterList::close();
					  exitscope();
					}
		;

declaration1_1  :	init_dcltrs
		|	proto_dcltr
		|	init_dcltrs ',' declaration1_2
		| 	proto_dcltr ',' declaration1_2
		| 	proto_dcltr ',' declaration1_3
		;

declaration1_2	:	proto_dcltr
		| 	proto_dcltr ',' declaration1_2
		| 	proto_dcltr ',' declaration1_3
		;

declaration1_3	:	init_dcltrs
		|	init_dcltrs ',' declaration1_2
		;

declaration2_1  :	init_dcltrs2
		|	proto_dcltr
		|	init_dcltrs2 ',' declaration2_2
		|	proto_dcltr  ',' declaration2_2
		|	proto_dcltr  ',' declaration2_3
		;

declaration2_2  :	proto_dcltr
		|	proto_dcltr ',' declaration2_2
		|	proto_dcltr ',' declaration2_3
		;

declaration2_3  :	init_dcltrs2
		|	init_dcltrs2 ',' declaration2_2
		;

var_decls	:       /* empty */
		|	var_decls var_declaration
		;

var_declaration	:	specs init_dcltrs2 ';'
                |       specs_notype init_dcltrs ';'
		|	specs ';'
                |       specs_notype ';'
		|	linkage ';'
		|	TYPEDEF type_spec { settypedeftype($2); } type_dcltrs ';'
                |	TYPEDEF type_spec ';' 			{ ; }
		;

linkage		:	EXTERN STRING_CONSTANT '{' decls '}' 	{ ; }
		|	EXTERN STRING_CONSTANT declaration 	{ ; }
		;

specs_notype	:	storage			{ settype("int");
						  setprototype(false);
                                                  setfunctype(gettype()); }
                ;

specs		:	storage type_spec	{ setfunctype(gettype()); }
		|	type_spec		{ setstorage(STORAGE_NONE);
						  setfunctype(gettype()); }
                |       type_spec storage       { setfunctype(gettype()); }
		;

storage		:	AUTO                    { setstorage(STORAGE_AUTO); }
                |	EXTERN                  { setstorage(STORAGE_EXTERN); }
		|	REGISTER		{ setstorage(STORAGE_REGISTER); }
		|	STATIC			{ setstorage(STORAGE_STATIC); }
		;

cast_spec	:	type_spec2		{ $$ = $1; newtagged($1); }

type_spec       :       type_spec2      	{ $$ = $1;
						  newtagged($1);
						  settype($1.QualifiedName());
						  setprototype($1.Prototype());
                                        	}

unORsigned	:	UNSIGNED
		|	SIGNED
		;

type_spec2	:	type_spec3
		|	float_spec
		|	enum_spec
		|	struct_spec
		|	union_spec
		|	DATATYPENAME
		|	DATATYPENAME type_spec3
		|	PROTOTYPENAME		      { $$.SetPrototype(true); }
		|	PROTOTYPENAME type_spec3      { $$.SetPrototype(true); }
		|	TYPEOF '(' exprs ')'	{ $$ = $3;
						  $$.SetName($$.Type()); }
		|	VOID
		;

type_spec3	:	int_spec
		|	unORsigned                   { 
				$$ = $1; 
				$$.SetName("int"); 
				$$.SetType(NULL); 
				$$.SetPrototype(false);
			}
		|	unORsigned int_spec          { $$ = $2; }
		|	int_spec unORsigned
		|	int_spec unORsigned int_spec { $$ = $3; }
		;

declarator	:	IDENTIFIER
		|	'(' declarator ')' parameter_list          { ParameterList::close();
                                                                     exitscope();
                                                                     $$ = $2; }
                |       declarator '[' exprs_opt ']'
                |       '*' declarator %prec Prefix	           { $$ = $2; }
 		|	'(' declarator ')'		           { $$ = $2; }
                ;

declarator2	:	id_or_name
		|	'(' declarator2 ')' parameter_list         { ParameterList::close();
                                                                     exitscope();
                                                                     $$ = $2; }
                |       declarator2 '[' exprs_opt ']'
                |       '*' declarator2 %prec Prefix	           { $$ = $2; }
 		|	'(' declarator2 ')'		           { $$ = $2; }
                ;

declarator3	:	id_or_name
		|	declarator3 parameter_list                 { ParameterList::close();
                                                                     exitscope();
                                                                     $$ = $1; 
								     $$.SetFunction(true); }
                |       declarator3 '[' exprs_opt ']'
                |       '*' declarator3 %prec Prefix	           { $$ = $2; $$.SetPointer(true); }
 		|	'(' declarator3 ')'		           { $$ = $2; }
                ;

parameter_list  :       '(' { enterscope(); ParameterList::start(); } parameters ')'
                ;

parameters      :       parameter
                |       parameters ',' parameter
                |       parameters ',' ELLIPSIS
                ;

parameter       :       parameter_type declarator3      {
                                                          $2.SetType(gettype());
							  addsym($2, false);
                                                          ParameterList::add($2);
                                                        }
                |       parameter_type unnamed_decl
                |       parameter_notype declarator     {
                                                          $2.SetType(gettype());
                                                          addsym($2, false);
                                                          ParameterList::add($2);
                                                        }
                |       parameter_notype unnamed_decl
                ;

parameter_type  :       type_spec			{ ; }
                |       REGISTER type_spec              { ; }
                |       type_spec REGISTER		{ ; }
                ;

parameter_notype:       /* empty */                     { settype("int"); 
							  setprototype(false); }
                |       REGISTER                        { settype("int"); 
							  setprototype(false); }
                ;

unnamed_decl    :       /* empty */
		|	'(' unnamed_decl ')' parameter_list     { ParameterList::close();
                                                                  exitscope(); }
                |       unnamed_decl '[' exprs_opt ']'
                |       '*' unnamed_decl %prec Prefix
                |	'(' unnamed_decl ')'		
                ;

init_dcltrs	:	declarator			{
                                                          $1.SetType(gettype());
							  $1.SetStorage(getstorage());
							  $1.SetPrototype(getprototype());
							  checkaddvar($1);
							  newvar($1);
                                                        }
		|	declarator Asgn_Op              {
                                                          $1.SetType(gettype());
							  $1.SetStorage(getstorage());
							  $1.SetPrototype(getprototype());
							  checkaddvar($1);
							  newvar($1);
                                                        } exprOrInit
		|	init_dcltrs ',' declarator2	{
                                                          $3.SetType(gettype());
							  $3.SetStorage(getstorage());
							  $3.SetPrototype(getprototype());
							  checkaddvar($3);
							  newvar($3);
							}
		|	init_dcltrs ',' declarator2 Asgn_Op {
                                                              $3.SetType(gettype());
							      $3.SetStorage(getstorage());
							      $3.SetPrototype(getprototype());
							      checkaddvar($3);
							      newvar($3);
                                                            } exprOrInit
		;

init_dcltrs2	:	declarator2			{
                                                          $1.SetType(gettype());
							  $1.SetStorage(getstorage());
							  $1.SetPrototype(getprototype());
							  checkaddvar($1);
							  newvar($1);
                                                        }
		|	declarator2 Asgn_Op              {
                                                          $1.SetType(gettype());
							  $1.SetStorage(getstorage());
							  $1.SetPrototype(getprototype());
							  checkaddvar($1);
							  newvar($1);
                                                        } exprOrInit
		|	init_dcltrs2 ',' declarator2	{
                                                          $3.SetType(gettype());
							  $3.SetStorage(getstorage());
							  $3.SetPrototype(getprototype());
							  checkaddvar($3);
							  newvar($3);
							}
		|	init_dcltrs2 ',' declarator2 Asgn_Op {
                                                               $3.SetType(gettype());
							       $3.SetStorage(getstorage());
							       $3.SetPrototype(getprototype());
							       checkaddvar($3);
							       newvar($3);
                                                             } exprOrInit
		;

type_dcltrs	:	declarator3
			{
			    const ParseInfo& info = gettypedeftype();
			    $1.SetType(info.QualifiedName());
			    newtype($1.Name(), $1.Prototype() ? 
					PROTOTYPENAME : DATATYPENAME);
			    addsym($1, false);
			    newtypedef($1);
			    $$ = $1;
			}
		|	type_dcltrs ',' declarator3
			{
			    const ParseInfo& info = gettypedeftype();
			    $3.SetType(info.QualifiedName());
			    newtype($3.Name(), $3.Prototype() ?
					PROTOTYPENAME : DATATYPENAME);
			    addsym($3, false);
			    newtypedef($3);
			    $$ = $3;
			}
		;


/*----------------------------   TYPES   ----------------------------------*/


int_spec	:	CHAR
		|	SHORT
		|	SHORT INT
		|	INT
		|	LONG INT
		|	LONG LONG
		|	LONG LONG INT
		|	LONG
		;
float_spec	:	FLOAT
		|	LONG FLOAT	{ $$ = $2; }
		|	DOUBLE
		|	LONG DOUBLE	{ $$ = $2; }
		;

enum_spec	:	ENUM id_or_name
			{
			    $$ = $2;
			    appendToName($$, "(enum)");
			    $$.SetType("int");
    			    const ParseInfo *temp = findsym($$.Name());
			    if (temp != NULL)
			      $$.SetQualifiedName(temp->QualifiedName());
			    else
			      addsym($$, false);
			}
		|	ENUM id_or_name 
			{
			    appendToName($2, "(enum)");
			    $2.SetType("int");
    			    const ParseInfo *temp = findsym($2.Name());
			    if (temp != NULL)
			      $2.SetQualifiedName(temp->QualifiedName());
			    else
			      addsym($2, false);
			    setstruct($2.QualifiedName());
			} 
			'{' enum_fields '}' 
			{
			    $$ = $2;
			    popstruct();
		            newtypedef($$);
			}
		|	ENUM 
			{
			    $1.MakeName();
			    appendToName($1, "(enum)");
			    $1.SetType("int");
			    addsym($1, false);
			    setstruct($1.QualifiedName());
			}
			'{' enum_fields '}'
			{
			    $$ = $1;
			    popstruct();
			    newtypedef($$);
			}
		;

struct_spec	:	STRUCT id_or_name
			{
			    $$ = $2;
			    appendToName($$, "(struct)");
			    $$.SetType(NULL);
    			    const ParseInfo *temp = findsym($$.Name());
			    if (temp != NULL)
                              $$.SetQualifiedName(temp->QualifiedName());
			    else
			      addsym($$, false);
			}
		|	STRUCT id_or_name
				{ appendToName($2, "(struct)");
				  $2.SetType(NULL);
				  setstruct($2.Name());
    			          const ParseInfo *temp = findsym($2.Name());
			          if (temp != NULL)
                                    $2.SetQualifiedName(temp->QualifiedName());
			          else
				    addsym($2, false); }
			'{' fields '}'
				{
				  $$ = $2;
				  popstruct();
				  newtypedef($$);
				}
		|	STRUCT	{ $1.MakeName();
				  appendToName($1, "(struct)");
				  setstruct($1.Name());
				  addsym($1, false); }
			'{' fields '}'
				{
				  $$ = $1;
				  popstruct();
				  newtypedef($$);
				}
		;

union_spec	:	UNION id_or_name
			{
			    $$ = $2;
			    appendToName($$, "(union)");
			    $$.SetType(NULL);
    			    const ParseInfo *temp = findsym($$.Name());
			    if (temp != NULL)
			      $$.SetQualifiedName(temp->QualifiedName());
			    else
			      addsym($$, false);
			}
		|	UNION id_or_name
				{ appendToName($2, "(union)");
				  $2.SetType(NULL);
				  setstruct($2.Name());
    			          const ParseInfo *temp = findsym($2.Name());
			          if (temp != NULL)
			            $2.SetQualifiedName(temp->QualifiedName());
			          else
				    addsym($2, false); }
			'{' fields '}'
				{
				  $$ = $2;
				  popstruct();
				  newtypedef($$);
				}
		|	UNION 	{ $1.MakeName();
				  appendToName($1, "(union)");
				  setstruct($1.Name());
				  addsym($1, false); }
			'{' fields '}'
				{
				  $$ = $1;
				  popstruct();
				  newtypedef($$);
				}
		;

id_or_name	:	IDENTIFIER
		|	DATATYPENAME
		|	PROTOTYPENAME		{ $$.SetPrototype(true); }
		;

fields		:	fields type_spec field_list ';'
		|	fields ';'
		|	error ';'
		|       /* empty */
		;

field_list	:	field
		|	field_list ',' field
		|	/* empty */
		;

field		:	declarator2             { addfield(getstruct(), $1.Name(), gettype()); }
		|	declarator2 ':' expr    { addfield(getstruct(), $1.Name(), gettype()); }
		|	':' expr
		;

enum_fields	:	enum_const
		|	enum_fields ',' enum_const
		;

enum_const	:	/* empty */
                |       id_or_name			{
							  settype(getstruct());
							  setprototype(false);
                                                          $1.SetType(gettype());
							  addsym($1, false);
							  newconst($1);
							}
		|	id_or_name Asgn_Op expr		{
							  settype(getstruct());
							  setprototype(false);
                                                          $1.SetType(gettype());
							  addsym($1, false);
							  newconst($1);
							}
		;

type_name	:	cast_spec 
		|	cast_spec abstract
		;

abstract	:	'(' abstract ')'
		|	'(' ')'
		|	abstract parameter_list {
				ParameterList::close();
                                exitscope(); 
			}
		|	'[' exprs_opt ']'
		|	abstract '[' exprs_opt ']'
		|	'*' %prec Prefix
		|	'*' abstract %prec Prefix
		;

/*--------------------------   STATEMENTS   --------------------------------*/


compound	:	'{' enterscope decls statements '}'	{ 
								  $$ = $4; 
								  exitscope(); 
								}
		|	error '}'			{ voidExpression($$); }
		;

enterscope	:					      { enterscope(); }
		;

statements	:				 	{ voidExpression($$); }
		|	statements statement_or_label		{ $$ = $2; }
		;

statement_or_label:     label ':'			{ voidExpression($$); }
                |       statement_no_label
                ;

statement       :       statement_no_label				{ ; }
		|	label ':' statement
                ;

statement_no_label:	expr_statement
		|	noexpr_statement		{ voidExpression($$); }

expr_statement	:	exprs ';'
		|	compound
		;

noexpr_statement:	IF '(' exprsOrCompound ')' statement		{ ; }
		|	IF '(' exprsOrCompound ')' statement ELSE statement {;}
		|	WHILE '(' exprsOrCompound ')' statement		{ ; }
		|	DO statement WHILE '(' exprsOrCompound ')' ';'	{ ; }
		|	FOR '(' exprs_opt ';' exprs_opt ';' exprs_opt ')'
			statement					{ ; }
		|	SWITCH '(' exprsOrCompound ')' statement	{ ; }
		|	BREAK ';'					{ ; }
		|	CONTINUE ';'					{ ; }
		|	RETURN exprs_opt ';'				{ ; }
		|	GOTO id_or_name ';'				{ ; }
		|	';'
		|	error ';'
		|       ASM ';'
		;

label		:	IDENTIFIER 	         			{ ; }
			/* this should be id_or_name (you can use a typename */
                        /* for a label), but that would cause a s/r conflict */
		|	CASE expr					{ ; }
		|	CASE expr ELLIPSIS expr				{ ; }
		|	DEFAULT						{ ; }
		;


/*---------------------------   EXPRESSIONS   -------------------------------*/


exprs_opt	:
		|	exprs		{ ; }
		|	compound	{ ; }
		;

exprOrInit	:       expr				
                |       '{' '}'				{ ; }
		|	'{' init_list '}'		{ $$ = $2; }
		|	'{' init_list ',' '}'		{ $$ = $2; }
		|	'[' expr ']' Asgn_Op exprOrInit { $$ = $5; }
		|	'[' expr ELLIPSIS expr ']' Asgn_Op exprOrInit 
							{ $$ = $7; }
		;

init_list	:	exprOrInit
		|	id_or_name ':' exprOrInit	{ $$ = $3; }
		|	init_list ',' exprOrInit
		|	init_list ',' id_or_name ':' exprOrInit
		;

exprsOrCompound	:	exprs           { ; }
                |	compound	{ ; }
		;

exprs		:	expr
		|	exprs ',' expr
		;

expr		:	comp_expr
		|	func_expr	{ $$ = $1; newcall($1); }
		|	IDENTIFIER
			{
			    $$ = $1;
			    newref($1);
			}
		;

comp_expr	: 	expr Asgn_Op expr
		|	expr '?' exprs_opt ':' expr 		{ $$ = $5; }
		|	expr Or expr
		|	expr And expr
		|	expr '|' expr
		|	expr '^' expr
		|	expr '&' expr
		|	expr Eq_Op expr
		|	expr Rel_Op expr
		|	expr Shift expr
		|	expr '+' expr { $$ = hasBuiltinType($3) ? $1 : $3; }
		|	expr '-' expr
		|	expr '*' expr
		|	expr '/' expr
		|	expr '%' expr
		|	expr IncDec
		|	expr '[' exprs ']'
		|	'*' expr %prec Prefix			{ $$ = $2; }
		|	'&' expr %prec Prefix			{ $$ = $2; }
		|	'+' expr %prec Prefix			{ $$ = $2; }
		|	'-' expr %prec Prefix			{ $$ = $2; }
		|	'!' expr %prec Prefix			{ $$ = $2; }
		|	'~' expr %prec Prefix			{ $$ = $2; }
                |       And expr %prec Prefix                   { $$ = $2; }
		|	IncDec expr %prec Prefix		{ $$ = $2; }
		|	SIZEOF '(' type_name ')' %prec SizeOf
			{
			    $$ = $1;
			    $$.SetName(NULL);
			    $$.SetType("int");
			}
		|	SIZEOF expr %prec SizeOf
			{
			    $$ = $1;
			    $$.SetName(NULL);
			    $$.SetType("int");
			}
		|	'(' exprs ')'
			{
			    $$ = $2;
			    $$.SetName(NULL);
			}
		|	'(' compound ')' 			{ $$ = $2; }
		|	constant
		|	'(' type_name ')' expr %prec Prefix
			{
			    $$ = $4;
			    $$.SetType($2.QualifiedName());
			}
		|	'(' type_name ')' '{' init_list '}' %prec Prefix
			{
			    $$ = $5;
			    $$.SetType($2.QualifiedName());
			}
		|	'(' type_name ')' '{' init_list ',' '}' %prec Prefix 
			{
			    $$ = $5;
			    $$.SetType($2.QualifiedName());
			}
		|	expr Arrow id_or_name
			{
			    newdata($1);
			    $$ = $1;
			    $$.SetType(fieldtype($1, $3.Name()));
			}
		|	expr '.' id_or_name
			{
			    newdata($1);
			    $$ = $1;
			    $$.SetType(fieldtype($1, $3.Name()));
			}
		;

func_expr	:	expr '(' exprs_opt ')'
		;

constant        :       CONSTANT
                |       string_constant
                ;

string_constant :       STRING_CONSTANT
                |       string_constant STRING_CONSTANT
                ;

%%
