/***********************************************************************/
/*                                                                     */
/* RIGI C Parser --- Copyright (C) 1996-1999                           */
/*                    The Rigi Group, University of Victoria           */
/*                                                                     */
/* see the file license.terms for usage and copying policies           */
/*                                                                     */
/* Maintainer: Johannes Martin (jmartin@csr.csc.uvic.ca)               */
/*                                                                     */
/***********************************************************************/
#include <stdio.h>
#include "cl_symtbl.h"
#include "cl_fldtbl.h"
#include "globals.h"
#include "clex.h"
#include "cinit.h"
#include "parse.h"

static const int MAX_SCOPE = 64;
static SymTbl symtbl[MAX_SCOPE];
static FldTbl fldtbl[MAX_SCOPE];
static int    scptbl[MAX_SCOPE];
static int scopeLevel = 0;
static int scopeCount = 0;

int enterscope() {
  if (scopeLevel >= MAX_SCOPE) {
    char buffer[80];
    sprintf(buffer, "too may scope levels (%d)", scopeLevel);
    yyerror(buffer);
    return 0;
  }
  if (scopeLevel++ == 0)
    scopeCount = 0;
  scptbl[scopeLevel] = scopeCount++;
  if (symdebug)
    fprintf(stderr, "enterscope: scope level = %d\n", scopeLevel);
  return 1;
}

int exitscope() {
  if (scopeLevel == 0) {
    yyerror("unbalanced scoping levels");
    return 0;
  }
  symtbl[scopeLevel].Empty();
  fldtbl[scopeLevel].Empty();
  scopeLevel--;
  if (symdebug)
    fprintf(stderr, "exitscope: scope level = %d\n", scopeLevel);
  return 1;
}

void print_symtbl() {
  int i;
  fprintf(stdout, "\nSymbol Table:\n");
  for(i=0; i<=scopeLevel ; i++) {
    fprintf(stdout, "Scope level %d\n", i);
    symtbl[i].Print();
  }
}

int addsym(ParseInfo &p, bool inTopScope) {
  if (!p.Name()) { 
    yyerror("internal error: addsymbol - NULL symbol"); 
    return 0; 
  }
  if (symdebug)
    fprintf(stderr, "addsymbol: %s %s at scope level %d\n",
	    p.Name(), p.Type() ? p.Type() : "NULL", scopeLevel);

  if (inTopScope || (scopeLevel == 0)) {
    char qualifiedName[strlen(p.Name()) + strlen(p.File()) + 15];
    sprintf(qualifiedName, "%s^%s^%d^%d", p.Name(), p.File(), p.Line(), p.Col());
    p.SetQualifiedName(qualifiedName);
    symtbl[0].SetSymbol(p);
  } else {
    const char *funcName = getfunc();
    char qualifiedName[strlen(p.Name()) + strlen(funcName) + 15];
    sprintf(qualifiedName, "%s^%s^%d", p.Name(), funcName, scptbl[scopeLevel]);
    p.SetQualifiedName(qualifiedName);
    symtbl[scopeLevel].SetSymbol(p);
  }
  //    if (symdebug) print_symtbl();
  return 1;
}

const ParseInfo* findsym(const char *key) {
  if (!key) {
    yyerror("internal error: findsym - NULL symbol");
    return NULL;
  }

  for(int i = scopeLevel; i >= 0; i--) {
    const ParseInfo *info = symtbl[i].GetSymbol(key);
    if (info)
      return info;
  }
  if (dowarn)
    fprintf(stderr,"warning: Could not find %s in the symbol table.\n",
	    key, scopeLevel);
  return NULL;
}

void addfield( const char* structure, const char* field, const char* type) {
    if (symdebug)
        fprintf(stderr,"addfield: structure = %s, field = %s, type = %s\n",
                structure, field, type);
    if (structure && field && type)
      fldtbl[scopeLevel].SetField(structure, field, type);
    else
      yyerror("internal error: addfield - couldn't set field");
}

const char* fieldtype( const ParseInfo& s, const char* f )
{
    if (!f) {
      yyerror("internal error: fieldtype - NULL field name");
      return NULL;
    }
    if (s.Type()) {
      const char *c = s.Type();
      char typeName[strlen(c) + 1];
      char *d = typeName;

      while ((*c != '^') && (*c != '\0'))
	*(d++) = *(c++);
      *d = '\0';
      const ParseInfo* t = findsym(typeName);
      if (t)
	return fieldtype(*t, f);
      else {
	char buffer[180];
	sprintf(buffer, 
		"unknown structured type %s, looking for field %s of %s.", 
		typeName, f, s.Name());
	yyerror(buffer);
	return NULL;
      }
    }
    for (int i = scopeLevel; i >= 0; i--) {
      const char *type = fldtbl[i].GetField(s.Name(), f);
      if (type) {
        if (symdebug)
          fprintf(stderr,"fieldtype: Field type for (%s,%s) = %s\n", s.Name(), f, type);
        return type;
      }
    }

    char buffer[80];
    sprintf(buffer, "no type for field %s of structure %s", f, s.Name());
    yyerror(buffer);
    //        print_symtbl();
    return NULL;
}

bool hasBuiltinType( const ParseInfo& s )
{
  if (s.Type()) {
    const char *c = s.Type();
    char typeName[strlen(c) + 1];
    char *d = typeName;
    
    while ((*c != '^') && (*c != '\0'))
      *(d++) = *(c++);
    *d = '\0';
    const ParseInfo *t = findsym(typeName);
    return t ? hasBuiltinType(*t) : true;
  }
  return (s.Name() && findsym(s.Name())) ? false : true;
}
